from http import HTTPStatus
from typing import Any, Awaitable, Callable, Tuple

from httpx import AsyncClient, Client, RequestError

__all__ = ("send_request", "SendRequestFn",)

SendRequestFn = Callable[[str, float, Any], Tuple[int, Any]]


class TelemetryRequestError(Exception):
    pass


def send_request(url: str, timeout: float, payload: Any) -> Tuple[int, Any]:
    with Client() as client:
        try:
            response = client.post(url, json=payload, timeout=timeout)
        except RequestError as e:
            raise TelemetryRequestError(f"Failed to send events to {url!r}: «{e!r}»") from None

        status = response.status_code
        try:
            body = response.json()
        except:  # noqa: E722
            body = response.text

    if status != HTTPStatus.OK:
        raise TelemetryRequestError(f"Failed to send events to {url!r}: {status} «{body}»")
    return status, body


SendRequestAsyncFn = Callable[[str, float, Any], Awaitable[Tuple[int, Any]]]


async def send_request_async(url: str, timeout: float, payload: Any) -> Tuple[int, Any]:
    async with AsyncClient() as client:
        try:
            response = await client.post(url, json=payload, timeout=timeout)
        except RequestError as e:
            raise TelemetryRequestError(f"Failed to send events to {url!r}: «{e!r}»") from None

        status = response.status_code
        try:
            body = response.json()
        except:  # noqa: E722
            body = response.text

    if status != HTTPStatus.OK:
        raise TelemetryRequestError(f"Failed to send events to {url!r}: {status} «{body}»")
    return status, body



from importlib.metadata import PackageNotFoundError, metadata
from pathlib import Path
from time import time
from typing import Optional, Union

__all__ = ("now", "get_project_name", "get_package_version",)


def _get_project_name(path: Path) -> Union[str, None]:
    maybe_git = path / ".git"
    if maybe_git.exists() and maybe_git.is_dir():
        return path.name

    if path == Path(path.root):
        return None

    return _get_project_name(path.parent)


def get_project_name(path: Optional[Path] = None, *, default: str = "") -> str:
    if path is None:
        path = Path()

    if not path.is_absolute():
        path = path.absolute()

    if project_name := _get_project_name(path):
        return project_name
    return default


def get_package_version(name: str, *, default: str = "0.0.0") -> str:
    try:
        version = metadata(name)
    except PackageNotFoundError:
        return default
    return str(version["Version"]) if ("Version" in version) else default


def now() -> int:
    return round(time() * 1000)



from abc import ABC, abstractmethod
from typing import Any, Dict, List, TypedDict, Union
from uuid import UUID

from ._utils import now

__all__ = ("StartedTelemetryEvent", "ArgParseTelemetryEvent", "ArgParsedTelemetryEvent",
           "StartupTelemetryEvent", "ExcRaisedTelemetryEvent", "EndedTelemetryEvent",
           "TelemetryEvent", "PluginInfo", "ExceptionInfo", "EnvironmentInfo",)


class PluginInfo(TypedDict):
    name: str
    module: str
    enabled: bool
    version: str


class ExceptionInfo(TypedDict):
    type: str
    message: str
    traceback: List[str]


class EnvironmentInfo(TypedDict):
    python_version: str
    vedro_version: str


class TelemetryEvent(ABC):
    def __init__(self) -> None:
        self._created_at = now()

    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        pass

    @abstractmethod
    def __repr__(self) -> str:
        pass


class StartedTelemetryEvent(TelemetryEvent):
    def __init__(self, session_id: UUID, project_id: str, inited_at: int,
                 environment: EnvironmentInfo, plugins: List[PluginInfo]) -> None:
        super().__init__()
        self._session_id = str(session_id)
        self._project_id = project_id
        self._inited_at = inited_at
        self._environment = environment
        self._plugins = plugins

    def to_dict(self) -> Dict[str, Any]:
        return {
            "event_id": f"{self.__class__.__name__}",
            "session_id": self._session_id,
            "created_at": self._created_at,
            "project_id": self._project_id,
            "inited_at": self._inited_at,
            "environment": self._environment,
            "plugins": self._plugins,
        }

    def __repr__(self) -> str:
        return (f"<{self.__class__.__name__} session_id={self._session_id!r} "
                f"project_id={self._project_id!r}>")


class ArgParseTelemetryEvent(TelemetryEvent):
    def __init__(self, session_id: UUID, cmd: List[str]) -> None:
        super().__init__()
        self._session_id = str(session_id)
        self._cmd = cmd

    def to_dict(self) -> Dict[str, Any]:
        return {
            "event_id": f"{self.__class__.__name__}",
            "session_id": self._session_id,
            "created_at": self._created_at,
            "cmd": self._cmd
        }

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} session_id={self._session_id!r} cmd={self._cmd!r}>"


class ArgParsedTelemetryEvent(TelemetryEvent):
    def __init__(self, session_id: UUID, args: Dict[str, Any]) -> None:
        super().__init__()
        self._session_id = str(session_id)
        self._args = args

    def to_dict(self) -> Dict[str, Any]:
        return {
            "event_id": f"{self.__class__.__name__}",
            "session_id": self._session_id,
            "created_at": self._created_at,
            "args": self._args
        }

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} session_id={self._session_id!r} args={self._args!r}>"


class StartupTelemetryEvent(TelemetryEvent):
    def __init__(self, session_id: UUID, discovered: int, scheduled: int) -> None:
        super().__init__()
        self._session_id = str(session_id)
        self._discovered = discovered
        self._scheduled = scheduled

    def to_dict(self) -> Dict[str, Any]:
        return {
            "event_id": f"{self.__class__.__name__}",
            "session_id": self._session_id,
            "created_at": self._created_at,
            "discovered": self._discovered,
            "scheduled": self._scheduled
        }

    def __repr__(self) -> str:
        return (f"<{self.__class__.__name__} session_id={self._session_id!r} "
                f"discovered={self._discovered} scheduled={self._scheduled!r}>")


class ExcRaisedTelemetryEvent(TelemetryEvent):
    def __init__(self, session_id: UUID, scenario_id: str, exception: ExceptionInfo) -> None:
        super().__init__()
        self._session_id = str(session_id)
        self._scenario_id = scenario_id
        self._exception = exception

    def to_dict(self) -> Dict[str, Any]:
        return {
            "event_id": f"{self.__class__.__name__}",
            "session_id": self._session_id,
            "created_at": self._created_at,
            "scenario_id": self._scenario_id,
            "exception": self._exception,
        }

    def __repr__(self) -> str:
        return (f"<{self.__class__.__name__} session_id={self._session_id!r} "
                f"scenario_id={self._scenario_id!r} exc_type={self._exception['type']!r}>")


class EndedTelemetryEvent(TelemetryEvent):
    def __init__(self, session_id: UUID, *, total: int, passed: int, failed: int,
                 skipped: int, interrupted: Union[ExceptionInfo, None]) -> None:
        super().__init__()
        self._session_id = str(session_id)
        self._total = total
        self._passed = passed
        self._failed = failed
        self._skipped = skipped
        self._interrupted = interrupted

    def to_dict(self) -> Dict[str, Any]:
        return {
            "event_id": f"{self.__class__.__name__}",
            "session_id": self._session_id,
            "created_at": self._created_at,
            "total": self._total,
            "passed": self._passed,
            "failed": self._failed,
            "skipped": self._skipped,
            "interrupted": self._interrupted,
        }

    def __repr__(self) -> str:
        is_interrupted = self._interrupted is not None
        return (f"<{self.__class__.__name__} session_id={self._session_id!r} "
                f"total={self._total} passed={self._passed} failed={self._failed} "
                f"skipped={self._skipped} is_interrupted={is_interrupted!r}>")



import atexit
import os
import sys
from argparse import ArgumentParser
from base64 import b64decode
from traceback import format_tb
from types import TracebackType
from typing import Any, List, Type, Union
from uuid import uuid4

from vedro.core import Dispatcher, ExcInfo, Plugin, PluginConfig
from vedro.events import (
    ArgParsedEvent,
    ArgParseEvent,
    CleanupEvent,
    ConfigLoadedEvent,
    ScenarioFailedEvent,
    StartupEvent,
)

from ._send_request import SendRequestFn, send_request
from ._utils import get_package_version, get_project_name, now
from .events import (
    ArgParsedTelemetryEvent,
    ArgParseTelemetryEvent,
    EndedTelemetryEvent,
    EnvironmentInfo,
    ExceptionInfo,
    ExcRaisedTelemetryEvent,
    PluginInfo,
    StartedTelemetryEvent,
    StartupTelemetryEvent,
    TelemetryEvent,
)

__all__ = ("VedroTelemetry", "VedroTelemetryPlugin",)


class VedroTelemetryPlugin(Plugin):
    _inited_at = now()

    def __init__(self, config: Type["VedroTelemetry"], *,
                 send_request: SendRequestFn = send_request) -> None:
        super().__init__(config)
        self._api_url = config.api_url.strip("/")
        self._timeout = config.timeout
        self._send_request = send_request
        self._session_id = uuid4()
        self._project_id = config.project_id or get_project_name(default="unknown")
        self._events: List[TelemetryEvent] = []
        self._arg_parser: Union[ArgumentParser, None] = None

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ConfigLoadedEvent, self.on_config_loaded) \
                  .listen(ArgParseEvent, self.on_arg_parse) \
                  .listen(ArgParsedEvent, self.on_arg_parsed) \
                  .listen(StartupEvent, self.on_startup) \
                  .listen(ScenarioFailedEvent, self.on_scenario_failed) \
                  .listen(CleanupEvent, self.on_cleanup)

    async def on_config_loaded(self, event: ConfigLoadedEvent) -> None:
        plugins: List[PluginInfo] = []
        for _, section in event.config.Plugins.items():
            name, module = section.plugin.__name__, section.plugin.__module__
            package = module.split(".")[0]
            if module.startswith("vedro.plugins") and section.enabled:
                continue
            plugins.append({
                "name": name,
                "module": module,
                "enabled": section.enabled,
                "version": get_package_version(package),
            })
        environment: EnvironmentInfo = {
            "python_version": sys.version,
            "vedro_version": get_package_version("vedro"),
        }
        self._events += [
            StartedTelemetryEvent(
                session_id=self._session_id,
                project_id=self._project_id,
                inited_at=self._inited_at,
                environment=environment,
                plugins=plugins,
            )
        ]
        atexit.register(self._send_events)

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        self._arg_parser = event.arg_parser
        path, *args = sys.argv
        prog = self._cleanup_arg(os.path.abspath(path))
        self._events += [ArgParseTelemetryEvent(self._session_id, [prog] + args)]

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        args = {}
        actions = getattr(self._arg_parser, "_actions", [])
        for action in actions:
            if action.dest not in event.args:
                continue
            arg = max(action.option_strings, key=len) if action.option_strings else action.dest
            val = getattr(event.args, action.dest)
            if val != action.default:
                args[arg] = self._cleanup_arg(val)

        self._events += [ArgParsedTelemetryEvent(self._session_id, args)]

    def on_startup(self, event: StartupEvent) -> None:
        discovered = len(list(event.scheduler.discovered))
        scheduled = len(list(event.scheduler.scheduled))
        self._events += [StartupTelemetryEvent(self._session_id, discovered, scheduled)]

    def on_scenario_failed(self, event: ScenarioFailedEvent) -> None:
        scenario_result = event.scenario_result
        # hacky way
        scenario_id = b64decode(scenario_result.scenario.unique_id + "===").decode()

        for step_result in scenario_result.step_results:
            exc_info = step_result.exc_info
            if exc_info is None:
                continue
            exception = self._format_exception(exc_info)
            self._events += [
                ExcRaisedTelemetryEvent(self._session_id, scenario_id, exception)
            ]

    def on_cleanup(self, event: CleanupEvent) -> None:
        report = event.report
        interrupted = self._format_exception(report.interrupted) if report.interrupted else None
        self._events += [
            EndedTelemetryEvent(
                session_id=self._session_id,
                total=report.total,
                passed=report.passed,
                failed=report.failed,
                skipped=report.skipped,
                interrupted=interrupted,
            )
        ]
        try:
            self._send_events()
        finally:
            atexit.unregister(self._send_events)

    def _send_events(self) -> None:
        payload = [e.to_dict() for e in self._events]
        self._send_request(f"{self._api_url}/v1/events", self._timeout, payload)
        self._events = []

    def _format_exception(self, exc_info: ExcInfo) -> ExceptionInfo:
        exc_type = exc_info.type
        return {
            "type": f"{exc_type.__module__}.{exc_type.__name__}",
            "message": str(exc_info.value),
            "traceback": self._format_traceback(exc_info.traceback),
        }

    def _format_traceback(self, tb: TracebackType) -> List[str]:
        return [self._cleanup_arg(x) for x in format_tb(tb, limit=100)]

    def _cleanup_arg(self, arg: Any, cwd: str = os.getcwd()) -> Any:
        if isinstance(arg, dict):
            return {k: self._cleanup_arg(v) for k, v in arg.items()}
        elif isinstance(arg, list):
            return [self._cleanup_arg(v) for v in arg]
        elif isinstance(arg, (type(None), bool, int, float)):
            return arg
        else:
            return str(arg).replace(cwd, ".")


class VedroTelemetry(PluginConfig):
    plugin = VedroTelemetryPlugin

    # Vedro Telemetry API URL
    api_url: str = "http://localhost:8080"

    # Timeout for requests to the API
    timeout: float = 5.0

    # Project ID
    project_id: Union[str, None] = None
